import textwrap
import types

import littletable as lt
from rich.text import Text
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import DataTable, Footer


class InteractiveLogMergeViewerApp(App):
    """
    Class to display merged results using textual TUI.
    """

    BINDINGS = [
        Binding(key="q", action="quit", description="Quit"),
    ]

    def config(
            self,
            log_file_names: list[str],
            display_width: int,
            show_line_numbers: bool,
            merged_log_lines_table: lt.Table,
    ):
        self.log_file_names = log_file_names  # noqa
        self.merged_log_lines_table = merged_log_lines_table  # noqa
        self.display_width = display_width  # noqa
        self.show_line_numbers = show_line_numbers  # noqa

    def compose(self) -> ComposeResult:
        yield DataTable()
        yield Footer()

    def on_mount(self) -> None:
        fixed_cols = 2 if self.show_line_numbers else 1
        col_names = self.merged_log_lines_table.info()["fields"]

        display_table = self.query_one(DataTable)
        display_table.cursor_type = "row"
        display_table.zebra_stripes = True
        display_table.fixed_columns = fixed_cols
        display_table.add_columns(*col_names)

        screen_width = self.display_width or self.size.width
        # guesstimate how much width each file will require
        width_per_file = int((screen_width - 25) * 0.94 // len(self.log_file_names))

        def max_line_count(sseq: list[str]):
            """
            The number of lines for this row is the maximum number of newlines
            in any value, plus 1.
            """
            return max(s.count("\n") for s in sseq) + 1

        line_ns: types.SimpleNamespace
        for line_ns in self.merged_log_lines_table:
            row_values = list(vars(line_ns).values())
            # see if any text wrapping is required for this line
            # - check each cell to see if any line in the cell exceeds width_per_file
            # - if not, just add this row to the display_table
            if any(len(rv_line) > width_per_file
                   for rv in row_values
                   for rv_line in rv.splitlines()):
                # wrap individual cells (except never wrap the timestamp or leading line number)
                wrapped_row_values = row_values[:fixed_cols]
                for cell_value in row_values[fixed_cols:]:
                    if "\n" in cell_value or len(cell_value) > width_per_file:
                        cell_lines = (
                            "\n".join(textwrap.wrap(rvl, width_per_file))
                            for rvl in cell_value.splitlines()
                        )
                        wrapped_row_values.append("\n".join(cell_lines))
                    else:
                        wrapped_row_values.append(cell_value)
                display_table.add_row(
                    Text(wrapped_row_values[0], justify="right")
                    if self.show_line_numbers else wrapped_row_values[0],
                    *wrapped_row_values[1:],
                    height=max_line_count(wrapped_row_values))
            else:
                display_table.add_row(
                    Text(row_values[0], justify="right")
                    if self.show_line_numbers else row_values[0],
                    *row_values[1:],
                    height=max_line_count(row_values))



from collections.abc import Sequence
import heapq
import itertools
from typing import TypeVar, Callable, Any

T = TypeVar("T")
KeyFunction = Callable[[T], Any]


class Merger:
    """
    Class that take a list of iterators and a key function, and yields each key value
    returned by calling that function on an item from that iterable and all the items from
    all the lists that have that value.

    Uses a heap to pull values from the multiple iterators, and itertools.groupby to combine
    values that are shared across multiple iterators by key value.
    """
    def __init__(self, seq_list: list[T], key_function: KeyFunction | None = None):
        self.seq_list = seq_list
        self.key_function = key_function or (lambda x: x)

        # compose an iterator that groups common entries returned from the heap that came from
        # different iterators
        self.heap_iter = itertools.groupby(
            heapq.merge(*self.seq_list, key=key_function),
            key=key_function
        )

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return next(self.heap_iter)
        except IndexError:
            raise StopIteration


if __name__ == '__main__':
    list1 = "ACEFG"
    list2 = "BCCDGKR"
    m = Merger([list1, list2])
    m = Merger([iter(list1), iter(list2)])
    m = Merger([(c for c in list1), (c for c in list2)])

    for key, items in m:
        print(list(items))

    label = lambda s, seq: ((s, obj) for obj in seq)

    m = Merger([label("1", list1), label("2", list2)], key_function=lambda x: x[1])
    for key, items in m:
        print(key, list(items))




from collections.abc import Generator, Callable, Iterable
from itertools import groupby, islice
from datetime import datetime
from typing import Any, Optional
from operator import itemgetter


class WindowedSort:
    def __init__(self, window: int, seq: Iterable, *, key: Optional[Callable] = None):
        self.seq = seq
        self.key_function = key

        self.seq_iter = iter(seq)
        self.lookahead_buffer = list(islice(self.seq_iter, window))
        self.lookahead_buffer.sort(key=self.key_function)
        self.seq_consumed = len(self.lookahead_buffer) < window

    def __iter__(self):
        return self

    def __next__(self):
        if not self.seq_consumed:
            try:
                self.lookahead_buffer.append(next(self.seq_iter))
                self.lookahead_buffer.sort(key=self.key_function)
            except StopIteration:
                self.seq_consumed = True

        if self.lookahead_buffer:
            return self.lookahead_buffer.pop(0)
        else:
            raise StopIteration()


class NewLogLineDetector:
    """
    Callable class used as a key function for itertools.groupby to detect log lines
    that don't start with a timestamp, and to group them with the last line that did
    have a timestamp.
    """
    def __init__(self):
        self._cur_dt = None

    def __call__(self, line_obj: tuple[datetime, Any]) -> datetime:
        dt, *_ = line_obj
        if dt is not None:
            self._cur_dt = dt
        return self._cur_dt


class MultilineLogCollapser:
    """
    Class to take an iterable of (datetime, str) tuples, and use itertools.groupby to
    merge consecutive log lines for a given datetime into one.

    Converts:
        2023-07-14 08:00:04 ERROR  Request processed unsuccessfully
        Something went wrong
        Traceback (last line is latest):
            blah
            blah
        ValueError("shouldn't have done that")
        2023-07-14 08:00:06 INFO   User authentication succeeded

    to two log entries.
    """
    def __init__(self):
        self._newlogline_detector = NewLogLineDetector()

    def __call__(self, log_seq: Iterable[tuple[datetime, str]]) -> Generator[tuple[datetime, str], None, None]:
        for timestamp, lines in WindowedSort(
                window=60,
                seq=((a, list(b)) for a, b in groupby(log_seq, key=self._newlogline_detector)),
                key=itemgetter(0)
        ):
            yield timestamp, "\n".join(line[1] for line in lines)


if __name__ == '__main__':
    from pathlib import Path
    from timestamp_wrapper import TimestampedLineTransformer
    log_lines = (Path("files") / "log1.txt").read_text().splitlines()
    transformer = TimestampedLineTransformer.make_transformer_from_sample_line(log_lines[0])

    for collapsed in MultilineLogCollapser()(transformer(line) for line in log_lines):
        print(collapsed)



from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path
import re

from typing import Callable, TypeVar, Union


T = TypeVar("T")
TimestampFormatter = Union[str | Callable[[str], datetime]]


class TimestampedLineTransformer:
    """
    Class to detect timestamp formats, and auto-transform lines that start with that timestamp into
    (timestamp, rest of the line) tuples.
    """
    pattern = ""
    match = lambda s: False

    def __init_subclass__(cls):
        cls.match = re.compile(cls.pattern).match

    @classmethod
    def make_transformer_from_file(cls, file_ref) -> TimestampedLineTransformer:
        if isinstance(file_ref, (str, Path)):
            with open(file_ref) as log_file:
                first_line = log_file.readline()
        else:
            first_line = file_ref.readline()

        xformer = cls.make_transformer_from_sample_line(first_line)
        xformer.add_file_info(Path(file_ref))
        return xformer

    @classmethod
    def make_transformer_from_sample_line(cls, s: str) -> TimestampedLineTransformer:
        for subcls in cls.__subclasses__():
            if subcls.match(s):
                return subcls()
        raise ValueError(f"no match for any timestamp pattern in {s!r}")

    def __init__(self, pattern: str, strptime_formatter: TimestampFormatter):
        self._re_pattern_match = re.compile(pattern).match
        self.pattern: str = pattern

        if isinstance(strptime_formatter, str):
            self.str_to_time = lambda s: datetime.strptime(s, strptime_formatter)
        else:
            self.str_to_time = strptime_formatter

        self.file_info: Path = None
        self.file_stat: os.stat_result = None

    def add_file_info(self, file_info: Path) -> None:
        self.file_info = file_info
        self.file_stat = file_info.stat()

    def __call__(self, obj: T) -> tuple[datetime | None, T]:
        m = self._re_pattern_match(obj)
        if m:
            # create (datetime, str) tuple - clip leading datetime string from
            # the log string, so that it doesn't duplicate when presented
            return self.str_to_time(m[1]), obj[m.span()[1]:]
        else:
            # no leading timestamp, just return None and the original string
            return None, obj


class YMDHMScommaF(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DD HH:MM:SS,SSS"
    pattern = r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3})\s"
    strptime_format = "%Y-%m-%d %H:%M:%S,%f"

    def __init__(self):
        super().__init__(self.pattern, self.strptime_format)


class YMDHMSdotF(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DD HH:MM:SS.SSS"
    pattern = r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3})\s"
    strptime_format = "%Y-%m-%d %H:%M:%S.%f"

    def __init__(self):
        super().__init__(self.pattern, self.strptime_format)


class YMDHMS(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DD HH:MM:SS"
    pattern = r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})\s"
    strptime_format = "%Y-%m-%d %H:%M:%S"

    def __init__(self):
        super().__init__(self.pattern, self.strptime_format)


class YMDTHMScommaF(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DDTHH:MM:SS,SSS"
    pattern = r"(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2},\d{3})\s"
    strptime_format = "%Y-%m-%dT%H:%M:%S,%f"

    def __init__(self):
        super().__init__(self.pattern, self.strptime_format)


class YMDTHMSdotF(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DDTHH:MM:SS.SSS"
    pattern = r"(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3})\s"
    strptime_format = "%Y-%m-%dT%H:%M:%S.%f"

    def __init__(self):
        super().__init__(self.pattern, self.strptime_format)


class YMDTHMS(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DDTHH:MM:SS"
    pattern = r"(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})\s"
    strptime_format = "%Y-%m-%d %H:%M:%S"

    def __init__(self):
        super().__init__(self.pattern, self.strptime_format)


class BDHMS(TimestampedLineTransformer):
    # syslog files with timestamp "mon day hh:mm:ss"
    # (note, year is omitted so let's guess from the log file's create date)
    pattern = r"([JFMASOND][a-z]{2}\s(\s|\d)\d \d{2}:\d{2}:\d{2})\s"
    strptime_format = "%b %d %H:%M:%S"

    def __init__(self):
        super().__init__(self.pattern, self.strptime_format)

    def __call__(self, obj: T) -> tuple[datetime | None, T]:
        # this format does not have a year so assume the file's create time year
        if self.file_stat is not None and self.file_stat.st_ctime:
            date_year = datetime.fromtimestamp(self.file_stat.st_ctime).year
        else:
            date_year = datetime.now().year

        dt, obj = super().__call__(obj)
        if dt is not None:
            dt = dt.replace(year=date_year)
        return dt, obj


class SecondsSinceEpoch(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DD HH:MM:SS"
    pattern = r"(\d{10})\s"

    def __init__(self):
        super().__init__(self.pattern, lambda s: datetime.fromtimestamp(int(s)))


class MilliSecondsSinceEpoch(TimestampedLineTransformer):
    # log files with timestamp "YYYY-MM-DD HH:MM:SS"
    pattern = r"(\d{13})\s"

    def __init__(self):
        super().__init__(self.pattern, lambda s: datetime.fromtimestamp(int(s) / 1000))


class FloatSecondsSinceEpoch(TimestampedLineTransformer):
    # log files with timestamp "1694561169.550987" or "1694561169.550"
    pattern = r"(\d{10}.\d+)\s"

    def __init__(self):
        super().__init__(self.pattern, lambda s: datetime.fromtimestamp(float(s)))


if __name__ == '__main__':
    files = "log1.txt log3.txt syslog1.txt".split()
    file_dir = Path(__file__).parent.parent / "files"
    xformers = [TimestampedLineTransformer.make_transformer_from_file(file_dir / f) for f in files]
    for fname, xformer in zip(files, xformers):
        log_lines = (file_dir / fname).read_text().splitlines()
        for ln in log_lines[:3]:
            print(xformer(ln))



#
# log_merger.py
#
# Utility for viewing multiple log files in a side-by-side merged format.
#
# Copyright 2023, Paul McGuire
#

import argparse
from collections.abc import Generator, Iterable
from datetime import datetime, timedelta
import itertools
import re
import sys
from typing import TypeVar

import littletable as lt

from .file_reading import FileReader
from .interactive_viewing import InteractiveLogMergeViewerApp
from .merging import Merger
from .multiline_log_handler import MultilineLogCollapser
from .timestamp_wrapper import TimestampedLineTransformer


T = TypeVar("T")


def make_argument_parser():
    epilog_notes = """
    Start and end timestamps to clip the given files to a particular time window can be 
    given in `YYYY-MM-DD HH:MM:SS.SSS` format, with trailing milliseconds and seconds
    optional, and "," permissible for the decimal point. A "T" can be included between
    the date and time to simplify entering the timestamp on a command line 
    (otherwise would require enclosing in quotes because of the intervening space). These
    command line values do not need to match the timestamp formats in the log files.

    These values may also be given as relative times, such as "15m" for "15 minutes ago".
    Valid units are "s", "m", "h", and "d" for seconds, minutes, hours, or days.
    """

    parser = argparse.ArgumentParser(epilog=epilog_notes)
    parser.add_argument("files", nargs="+", help="log files to be merged")
    parser.add_argument(
        "--interactive", "-i",
        action="store_true",
        help="show output using interactive TUI browser"
    )
    parser.add_argument('--start', '-s', required=False, help="start time to select time window for merging logs")
    parser.add_argument('--end', '-e', required=False, help="end time to select time window for merging logs")
    parser.add_argument(
        "--width", "-w",
        type=int,
        help="total screen width to use for interactive mode (defaults to current screen width)",
        default=0
    )
    parser.add_argument("--line_numbers", "-ln", action="store_true", help="add line number column")
    parser.add_argument("--csv", "-csv", help="save merged logs to CSV file")
    parser.add_argument(
        "--encoding", "-enc",
        type=str,
        default=sys.getfilesystemencoding(),
        help="encoding to use when reading log files (defaults to the system default encoding)")

    return parser


def parse_time_using(ts_str: str, formats: str | list[str]) -> datetime:
    if not isinstance(formats, (list, tuple)):
        formats = [formats]
    for fmt in formats:
        try:
            return datetime.strptime(ts_str, fmt)
        except ValueError:
            pass
    raise ValueError(f"no matching format for input string {ts_str!r}")


def parse_relative_time(ts_str: str) -> datetime:
    parts = re.match(r"(\d+)([smhd])$", ts_str, flags=re.IGNORECASE)
    if parts:
        qty, unit = parts.groups()
        seconds = int(qty)
        now = datetime.now()
        for unit_type, mult in [("s", 1), ("m", 60), ("h", 60), ("d", 24)]:
            seconds *= mult
            if unit == unit_type:
                return now - timedelta(seconds=seconds)

    raise ValueError(f"invalid relative time string {ts_str!r}")


def label(s: str, seq: Iterable[T]) -> Generator[tuple[str, T], None, None]:
    """
    method to make each item of an Iterable into a tuple containing the
    label (so that as items from different iterators are later combined, we'll know
    which iterator a particular item came from)
    """
    yield from ((s, obj) for obj in seq)


class LogMergerApplication:
    def __init__(self, config: argparse.Namespace):
        self.config = config

        self.fnames = config.files
        self.total_width = config.width

        valid_formats = [
            "%Y-%m-%d %H:%M:%S.%f",
            "%Y-%m-%d %H:%M:%S,%f",
            "%Y-%m-%d %H:%M:%S",
            "%Y-%m-%d %H:%M",
            "%Y-%m-%dT%H:%M:%S.%f",
            "%Y-%m-%dT%H:%M:%S,%f",
            "%Y-%m-%dT%H:%M:%S",
            "%Y-%m-%dT%H:%M",
            "%Y-%m-%d",
        ]
        if config.start is None:
            self.start_time = datetime.min
        else:
            if config.start.endswith(tuple("smhd")):
                self.start_time = parse_relative_time(config.start)
            else:
                self.start_time = parse_time_using(config.start, valid_formats)

        if config.end is None:
            self.end_time = datetime.max
        else:
            if config.end.endswith(tuple("smhd")):
                self.end_time = parse_relative_time(config.end)
            else:
                self.end_time = parse_time_using(config.end, valid_formats)

        if self.end_time <= self.start_time:
            raise ValueError("invalid start/end times - start must be before end")

        self.time_clip = lambda ts_log: self.start_time <= ts_log[0] <= self.end_time

        self.interactive = config.interactive
        self.textual_output = self.interactive
        self.table_output = not self.interactive
        self.save_to_csv = config.csv

        self.encoding = self.config.encoding

    def run(self):
        # generate dicts, one per timestamp, with values for each log file for the respective
        # log line from that file at that timestamp, or "" if no log line at that timestamp
        merged_lines = self._merge_log_file_lines()

        # build a littletable Table for easy tabular output, and insert the dicts of merged lines
        merged_lines_table = lt.Table()
        merged_lines_table.insert_many(merged_lines)

        if self.save_to_csv:
            merged_lines_table.csv_export(self.save_to_csv)

        elif self.table_output:
            # present the table - using a rich Table, the columns will auto-size to content and terminal
            # width
            merged_lines_table.present()

        elif self.textual_output:
            self._display_merged_lines_interactively(merged_lines_table)

    def _merge_log_file_lines(self) -> Generator[dict[str, T], None, None]:

        # scan input files to determine timestamp format, and create appropriate transformer for each
        readers = [FileReader.get_reader(fname, self.encoding) for fname in self.fnames]
        peek_iters, readers = zip(*[itertools.tee(rdr) for rdr in readers])
        transformers = [TimestampedLineTransformer.make_transformer_from_sample_line(next(peek_iter))
                        for peek_iter in peek_iters]

        # build iterators over each file that:
        # - transform each line into a (datetime, str) tuple (where the str is everything after the
        #   timestamp, so that it doesn't get repeated in the output table)
        # - collapses multiline logs (lines that go beyond just one line, with subsequent lines that
        #   do not start with a timestamp - tracebacks are a common example)
        # - labels each item the its source filename (so that after pulling an entry for the heap, we
        #   know which file it came from)
        # (for background on why we must use map() instead of a generator expression,
        # see https://chat.stackoverflow.com/transcript/message/56645472#56645472)
        log_file_line_iters = [
            (
                label(
                    fname,
                    filter(self.time_clip, (
                            MultilineLogCollapser()(
                                map(xformer, map(str.rstrip, reader))
                            )
                        )
                    )
                )
            ) for fname, xformer, reader in zip(self.fnames, transformers, readers)
        ]

        # use the Merger class which internally uses a heap to pull values in timestamp order from
        # all the iterators, and then uses groupby to group them by common timestamp
        merger = Merger(log_file_line_iters, key_function=lambda log_data: log_data[1][0])

        if self.config.line_numbers:
            initialize_row_dict = lambda n, ts: {"line": str(n), "timestamp": ts.strftime("%Y-%m-%d %H:%M:%S.%f")[:23]}
        else:
            initialize_row_dict = lambda n, ts: {"timestamp": ts.strftime("%Y-%m-%d %H:%M:%S.%f")[:23]}

        # build and yield a dict for each timestamp
        for line_number, (timestamp, items) in enumerate(merger, start=1):

            # initialize the entry for this timestamp with empty strings for each given file
            line_dict = {
                **initialize_row_dict(line_number, timestamp),
                **{}.fromkeys(self.fnames, ""),
            }

            # copy from the group each file's respective logging for this timestamp, and
            # insert into the dict for this timestamp
            for item in items:
                fname, (_, line) = item
                line_dict[fname] = line

            # yield the populated dict
            yield line_dict

    def _display_merged_lines_interactively(
            self,
            merged_log_lines: lt.Table,
    ):

        app = InteractiveLogMergeViewerApp()
        app.config(self.fnames, self.total_width, self.config.line_numbers, merged_log_lines)
        app.run()


def main():

    parser = make_argument_parser()
    args_ns = parser.parse_args()

    app = LogMergerApplication(args_ns)
    app.run()


if __name__ == '__main__':
    main()



from __future__ import annotations

import abc
import types
import typing


class FileReader:
    @classmethod
    def get_reader(cls, name: str, encoding: str) -> FileReader:
        for subcls in cls.__subclasses__():
            if subcls is TextFileReader:
                continue
            if subcls._can_read(name):
                return subcls(name, encoding)
        return TextFileReader(name, encoding)

    @classmethod
    @abc.abstractmethod
    def _can_read(cls, fname: str) -> bool:
        """Override in subclasses"""

    @abc.abstractmethod
    def _close_reader(self):
        """Override in subclasses"""

    def __init__(self, file_name: str, encoding: str):
        self.file_name = file_name
        self.encoding = encoding
        self._iter = iter(())

    def __iter__(self):
        return self._iter

    def __next__(self):
        try:
            return next(self._iter)
        except StopIteration:
            self._close_reader()
            raise


class TextFileReader(FileReader):
    @classmethod
    def _can_read(cls, fname: str) -> bool:
        return True

    def __init__(self, fname: str, encoding: str):
        super().__init__(fname, encoding)
        self._close_obj = open(self.file_name, encoding=self.encoding)
        self._iter = self._close_obj

    def _close_reader(self):
        self._close_obj.close()


class GzipFileReader(FileReader):
    @classmethod
    def _can_read(cls, fname: str) -> bool:
        return fname.endswith(".gz")

    def __init__(self, fname: str, encoding: str):
        import gzip

        super().__init__(fname, encoding)
        self._close_obj = gzip.GzipFile(filename=self.file_name)
        self._iter = (s.decode(self.encoding) for s in self._close_obj)
        # make fake stat result
        self.file_stat = types.SimpleNamespace(st_ctime=self._close_obj.mtime)

    def _close_reader(self):
        self._close_obj.close()
